﻿unit Cinema.Configuration;

interface

type

  { TCinemaConfiguration }

  TCinemaConfiguration = class
  private
    class var FDtlWorkDirScens: UnicodeString;
    class var FDtlWorkDir: UnicodeString;
    class var FCinemaPath: UnicodeString;
    class var FPlatformConfig: UnicodeString;
    class var FDtlPath: UnicodeString;
    class var FDtlAppPath: UnicodeString;
    class var FDtlAppPathPlugins: UnicodeString;
    class var FRtlPath: UnicodeString;
    class var FRtlAppPath: UnicodeString;
    class var FRtlCorePath: UnicodeString;
    class var FRtlScenariosPath: UnicodeString;
    class var FRtlPluginsDictionaryPath: UnicodeString;
    class var FRtlPluginsPath: UnicodeString;
    class var FRtlRelativeDtlPath: UnicodeString;
    class var FRtlRelativeDtlScenarioPath: UnicodeString;
    class constructor Create;
  public
    class procedure Save;
    class procedure Load;
    /// <summary> Синхронная очередь между потоками; если равна True тогда в случае полного заполнения очереди, пишущий поток
    /// будет ожидать её освобождения для записи очередного эдемента очереди </summary>
    class var SynchroQueues: boolean;
    /// <summary> Таймаут ожидания ответа из сценария/(процесса его исполняющего), мс </summary>
    class var TimeoutScenario: int32;
    /// <summary> Таймаут элемента сценария/(процесса его исполняющего) для перевода его в состояние TNodeState.nsNoData, мс </summary>
    class var TimeoutForStateNoData: int32;

    /// <summary> Полный путь до Cinema </summary>
    class property CinemaPath: UnicodeString read FCinemaPath;
    /// <summary> Подкаталоги $(Platform)$(Config) в пути к ядру синемы </summary>
    class property PlatformConfig: UnicodeString read FPlatformConfig;
    /// <summary> Полный путь до Cinema\dtl </summary>
    class property DtlPath: UnicodeString read FDtlPath;
    /// <summary> Полный путь до каталога приложений dtl. Учитывается есть ли платформа и конфигурация у текущего запускаемого </summary>
    class property DtlAppPath: UnicodeString read FDtlAppPath;
    /// <summary> Полный путь до каталога приложений dtl. Учитывается есть ли платформа и конфигурация у текущего запускаемого </summary>
    class property DtlWorkDir: UnicodeString read FDtlWorkDir;
    /// <summary> Полный путь до каталога приложений dtl. Учитывается есть ли платформа и конфигурация у текущего запускаемого </summary>
    class property DtlWorkDirScens: UnicodeString read FDtlWorkDirScens;
    /// <summary> Полный путь до каталога кастомных приложений-конфигураторов плагинов. Учитывается есть ли платформа и конфигурация у текущего запускаемого </summary>
    class property DtlAppPathPlugins: UnicodeString read FDtlAppPathPlugins;
    /// <summary> Полный путь до Cinema\rtl </summary>
    class property RtlPath: UnicodeString read FRtlPath;
    /// <summary> Полный путь до приложений rtl </summary>
    class property RtlAppPath: UnicodeString read FRtlAppPath;
    /// <summary> Полный путь до ядра Cinema. Учитывается есть ли платформа и конфигурация у текущего запускаемого </summary>
    class property RtlCorePath: UnicodeString read FRtlCorePath;
    /// <summary> Полный путь до каталога со сценариями </summary>
    class property RtlScenariosPath: UnicodeString read FRtlScenariosPath;
    /// <summary> Полный путь до файла со словарём зарегистрированных плагинов </summary>
    class property RtlPluginsDictionaryPath: UnicodeString read FRtlPluginsDictionaryPath;
    /// <summary> Полный путь до каталога с плагинами, учитывается есть ли платформа и конфигурация у текущего запускаемого </summary>
    class property RtlPluginsPath: UnicodeString read FRtlPluginsPath;
    /// <summary> Путь до rtl относительно dtl, учитывается есть ли платформа и конфигурация у текущего запускаемого </summary>
    class property RtlRelativeDtlPath: UnicodeString read FRtlRelativeDtlPath;
    /// <summary> Путь до каталога со сценариями относительно пути до каталоа приложений dtl </summary>
    class property RtlRelativeDtlScenarioPath: UnicodeString read FRtlRelativeDtlScenarioPath;
  end;

implementation

uses
    SysUtils
  , Cinema.Types
  , Cinema.Utils
  , Cinema.Exceptions
  ;

{ TCinemaConfiguration }

class constructor TCinemaConfiguration.Create;
const
  csDtlSubfolder = PathDelim + 'dtl' + PathDelim;
  csRtlSubfolder = PathDelim + 'rtl' + PathDelim;

  procedure BuildAsRtl(const AAppPath: UnicodeString; ARtlSubfolderIndex: Int32);
  begin
    FCinemaPath := Copy(AAppPath, 1, ARtlSubfolderIndex);
    FRtlPath := FCinemaPath.TrimRight([PathDelim]) + csRtlSubfolder;
    FPlatformConfig := UnicodeString(ExtractFilePath(Copy(AAppPath, length(FRtlPath), length(AAppPath) - length(FRtlPath)))).Trim([PathDelim]);
    if FPlatformConfig <> '' then
      FRtlCorePath := FRtlPath + FPlatformConfig + PathDelim + csCinemaCore
    else
      FRtlCorePath := FRtlPath + csCinemaCore;
    FRtlAppPath := ExtractFilePath(AAppPath);
    FDtlPath := Copy(AAppPath, 1, ARtlSubfolderIndex) + csDtlSubfolder;
    // обозначим для rtl бесполезную информацию
    FDtlAppPath := '';
    FDtlPath := '';
    FRtlRelativeDtlPath := '';
  end;

  procedure BuildAsDtl(const AAppPath: UnicodeString);
  var
    index: Int32;
    appSubPath: UnicodeString;
    appBeforeSubfolder: UnicodeString;
    appPathOther: UnicodeString;
    appPathCinema: UnicodeString;
    appPath: UnicodeString;
  begin
    index := AAppPath.IndexOf(csDtlSubfolder);
    appPath := AAppPath;
    if index <= 0 then
    begin
      appPathOther := AAppPath;
      appSubPath := '';
      appBeforeSubfolder := '';
      // dtl каталога может не быть, а rtl будет всегда, поэтому опираемся на него
      while (length(appPathOther) > length('Cinema' + csRtlSubfolder) + 1) do
      begin
        appPathOther := ExtractFilePath(appPathOther).TrimRight([PathDelim]);
        appPathCinema := IncludeTrailingPathDelimiter(appPathOther) + 'Cinema';
        if DirectoryExists(appPathCinema + csRtlSubfolder) then
        begin
          appPathCinema := appPathCinema + csDtlSubfolder;
          if (appBeforeSubfolder <> '') then
            appPath := appPathCinema + appSubPath.TrimLeft(IncludeTrailingPathDelimiter(appBeforeSubfolder)) + WideString('Hypervisor.exe')
          else
            appPath := appPathCinema + WideString('Hypervisor.exe');
          index := appPath.IndexOf(csDtlSubfolder);
          break;
        end;
        appBeforeSubfolder := ExtractFileName(appPathOther);
        appSubPath := IncludeTrailingPathDelimiter(appBeforeSubfolder) + appSubPath;
      end;
      if index <= 0 then
        raise ECinema.Create('Неизвестная структура каталогов!');
    end;

    FCinemaPath := Copy(appPath, 1, index);
    FDtlAppPath := ExtractFilePath(appPath);
    FDtlAppPathPlugins := IncludeTrailingPathDelimiter(FDtlAppPath + csPluginsDir);

    FDtlPath := ExcludeTrailingPathDelimiter(FCinemaPath) + csDtlSubfolder;
    FRtlPath := ExcludeTrailingPathDelimiter(Copy(appPath, 1, index)) + csRtlSubfolder;
    FPlatformConfig := UnicodeString(ExtractFilePath(Copy(appPath, length(FDtlPath), length(appPath) - length(FDtlPath)))).Trim([PathDelim]);
    if FPlatformConfig <> '' then
      FRtlAppPath := FRtlPath + FPlatformConfig + PathDelim
    else
      FRtlAppPath := FRtlPath;

    FRtlCorePath := FRtlAppPath + csCinemaCore;

    appPath := FDtlAppPath.TrimRight([PathDelim]);
    FRtlRelativeDtlPath := '';
    while length(appPath) > 3 do
    begin
      if FRtlPath.Contains(appPath) then
      begin
        FRtlRelativeDtlPath := FRtlRelativeDtlPath.TrimRight([PathDelim]) + csRtlSubfolder;
        break;
      end else
      begin
        FRtlRelativeDtlPath := '..' + PathDelim + FRtlRelativeDtlPath;
        appPath := ExtractFilePath(appPath.TrimRight([PathDelim]));
      end;
    end;
  end;

var
  app: UnicodeString;
  index: Int32;

begin
  // значения по умолчанию
  SynchroQueues := True;
  TimeoutScenario := 1000*30;
  TimeoutForStateNoData := 5000;

  app := WideString(ParamStr(0));
  index := app.IndexOf(csRtlSubfolder);
  if index > 0 then // запуск rtl
  begin
    BuildAsRtl(app, index);
  end else
  begin // запуск dtl
    BuildAsDtl(app);
    if not FileExists(FRtlCorePath) then
    begin // неудача; выйдем за пределы каталога программы и попробуем снова
      BuildAsDtl(ExtractFilePath(ExtractFilePath(app).TrimRight([PathDelim])));
      if not FileExists(FRtlCorePath) then
        raise ECinema.Create('Неизвестная структура каталогов!');
    end;
  end;

  FRtlPluginsDictionaryPath := FRtlPath + 'Plugins.json';
  FRtlScenariosPath := IncludeTrailingPathDelimiter(FRtlPath + csScenarioDir);
  FRtlPluginsPath := IncludeTrailingPathDelimiter(ExtractFilePath(FRtlCorePath) + csPluginsDir);
  FRtlRelativeDtlScenarioPath := IncludeTrailingPathDelimiter(FRtlRelativeDtlPath + csScenarioDir);
  FDtlWorkDir := FDtlAppPath + Cinema.Types.csWorkDir;
  FDtlWorkDirScens := FDtlWorkDir + Cinema.Types.csWorkSubDirScens;
  if not DirectoryExists(FRtlPluginsPath) then
    ForceDirectories(FRtlPluginsPath);
  if not DirectoryExists(FRtlScenariosPath) then
    ForceDirectories(FRtlScenariosPath);
end;

class procedure TCinemaConfiguration.Load;
begin
  // TODO
end;

class procedure TCinemaConfiguration.Save;
begin
  // TODO
end;

end.
