﻿unit Cinema.Scenario.Interfaces;


{$ifdef FPC}
  {.$mode delphi}
	{$mode delphiunicode}
  // Задан глобально в настройках проекта для платформы Win32. Заменяет все результаты функций с типом Interface на тип Pointer.
  {.$define NARI} // not allow return interfaces
{$endif}

interface

// TODO: вынести хит-тест, расчёт связи, выбор объекта по координатам и т.п. во вью (он же у меня контроллер)

uses
  {$ifdef FPC}
  fpjson,
  {$else}
  System.JSON,
  {$endif}
  Types,
  Cinema.Types,
  Cinema.Interfaces,
  Cinema.Log.Interfaces
  ;

type
  // TODO - перевести на абстрактные классы
  IScenarioNode = interface;
  IScenarioLayer = interface;
  IDataStreamProcessor = interface;

  TNodeState = (nsNone, nsStarting, nsWorking, nsNoData, nsStopping);
  TNodeStates = set of TNodeState;
  TNodeLogState = (lsInfo, lsWarn, lsError);
  TNodeLogStateSet = set of TNodeLogState;

  // 0 - очень плохо, 100 - всё хорошо
  TNodeHalthDegreeState = byte;

  TConnectionAlign = (caLeft, caRight, caTop, caBottom);

  TScenarioNodeEventNotify = procedure(const ASource: IScenarioNode) of object; stdcall;

  /// <summary> Связь между двумя элементами способными обрабатывать/транслировать/получать данные </summary>
  IScenarioConnection = interface
  ['{0B9C2AA7-C75F-46E4-BEAF-352F08B3BBC3}']
    function GetSource: {$ifdef NARI} Pointer {$else} IDataStreamProcessor; {$endif} stdcall;
    procedure SetSource(const AValue: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}); stdcall;
    function GetDestination: {$ifdef NARI} Pointer {$else} IDataStreamProcessor; {$endif} stdcall;
    procedure SetDestination(const AValue: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif}); stdcall;
    function GetNumber: Integer; stdcall;
    procedure SetNumber(const AValue: Integer); stdcall;
    function GetNumberInput: Integer; stdcall;
    procedure SetNumberInput(const AValue: Integer); stdcall;
    function GetQueue: Pointer; stdcall;
    procedure SetQueue(const AValue: Pointer); stdcall;
    function Save: PWideChar; stdcall;
    function Load(AValues: PWideChar; const AParentLayer: IScenarioLayer): boolean; stdcall;
    function GetView: TObject; stdcall;
    procedure SetView(const AValue: TObject); stdcall;
    function HitTest(X, Y: Integer): boolean; stdcall;
    procedure Calculate; stdcall;
    function GetAlignFrom: TConnectionAlign; stdcall;
    function GetAlignTo: TConnectionAlign; stdcall;
    function GetPointDst: PPoint; stdcall;
    function GetPointSrc: PPoint; stdcall;
    function GetRect: PRect; stdcall;
    function GetNameOutput: PWideChar; stdcall;
    function GetNameInput: PWideChar; stdcall;

    /// <summary> откуда </summary>
    property Source: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif} read GetSource write SetSource;
    /// <summary> куда </summary>
    property Destination: {$ifdef NARI} Pointer {$else} IDataStreamProcessor {$endif} read GetDestination write SetDestination;
    /// <summary> номер/индекс связи в Source </summary>
    property Number: Integer read GetNumber write SetNumber;
    /// <summary> номер входа в Destination </summary>
    property NumberInput: Integer read GetNumberInput write SetNumberInput;
    /// <summary> очередь для записи данных от Source елси Destination расположен в другом рабочем потоке </summary>
    property Queue: Pointer read GetQueue write SetQueue;
    /// <summary> указывает на сторону от которой выходит связь </summary>
    property AlignFrom: TConnectionAlign read GetAlignFrom;
    /// <summary> указывает на сторону в которую входит связь </summary>
    property AlignTo: TConnectionAlign read GetAlignTo;
    /// <summary> покрываемая область в которую вписана связь </summary>
    property Rect: PRect read GetRect;
    /// <summary> точка выхода из Source </summary>
    property PointSrc: PPoint read GetPointSrc;
    /// <summary> точка входа в Destination </summary>
    property PointDst: PPoint read GetPointDst;
    /// <summary> имя выхода </summary>
    property NameOutput: PWideChar read GetNameOutput;
    /// <summary> имя входа </summary>
    property NameInput: PWideChar read GetNameInput;
    /// <summary> вью связи </summary>
    property View: TObject read GetView write SetView;
  end;

  /// <summary> Базовый элемент сценария </summary>
  IScenarioNode = interface
  ['{5F8295B1-91A9-420F-A694-7B1CBC58ECC1}']
    function Clone(const AResult: IScenarioNode): {$ifdef NARI} Pointer {$else} IScenarioNode; {$endif} stdcall;
    function GetCaption: PWideChar; stdcall;
    procedure SetCaption(const ACaption: PWideChar); stdcall;
    function GetParent: {$ifdef NARI} Pointer {$else} IScenarioLayer {$endif}; stdcall;
    procedure SetParent(const AValue: {$ifdef NARI} Pointer {$else} IScenarioLayer {$endif}); stdcall;
    function GetGuid: System.PGuid; stdcall;
    procedure SetGuid(AValue: System.PGuid); stdcall;
    function GetState: TNodeState; stdcall;
    procedure SetState(const AValue: TNodeState); stdcall;
    function GetHealthState: TNodeHalthDegreeState; stdcall;
    function GetHealthStateMessage: PWideChar; stdcall;
    /// <summary> Смена текущего состояния качества выполняемой работы;
    /// название метода с префиксом Change вместо Set из-за того что уже есть
    /// метод ICinema.SetHealthState  </summary>
    procedure ChangeHealthState(AMsg: PWideChar; AHealthState: TNodeHalthDegreeState); stdcall;
    function GetLogState: TNodeLogState; stdcall;
    procedure SendNodeMessage(AMsg: PWideChar; AMessageType: TMessageType); stdcall;
    procedure SetLogState(AMessage: PWideChar; AValue: TNodeLogState); stdcall;
    /// <summary> Установка статуса лога в состояние "Без ошибок" </summary>
    procedure ResetLogState; stdcall;
    function GetLogStateMessage: PWideChar; stdcall;
    function GetThreadLayer: {$ifdef NARI} Pointer {$else} IScenarioLayer; {$endif} stdcall;
    function GetPosition: PPoint; stdcall;
    procedure SetPosition(AValue: PPoint); stdcall;
    function Save: PWideChar; stdcall;
    function Load(AValues: PWideChar; const AParentLayer: IScenarioLayer): boolean; stdcall;
    function GetHeight: Integer; stdcall;
    function GetWidth: Integer; stdcall;
    procedure Resize(AWidth, AHeight: Integer); stdcall;
    function GetView: TObject; stdcall;
    procedure SetView(const AValue: TObject); stdcall;
    function GetRect: PRect; stdcall;
    function GetNodeKey: PWideChar; stdcall;

    property Caption: PWideChar read GetCaption write SetCaption;
    property Parent: {$ifdef NARI} Pointer {$else} IScenarioLayer {$endif} read GetParent write SetParent;
    property Guid: PGuid read GetGuid write SetGuid;
    /// <summary> Рабочее состояние </summary>
    property State: TNodeState read GetState write SetState;
    /// <summary> Состояние лога; характеризует наличие наихудшего типа сообщения </summary>
    property LogState: TNodeLogState read GetLogState;
    property LogStateMessage: PWideChar read GetLogStateMessage;
    /// <summary> Текущее состояние качества выполняемой работы </summary>
    property HealthState: TNodeHalthDegreeState read GetHealthState;
    property HealthStateMessage: PWideChar read GetHealthStateMessage;
    property Position: PPoint read GetPosition write SetPosition;
    property Width: Integer read GetWidth;
    property Height: Integer read GetHeight;
    /// <summary> Текущая область занимаемая объектом в пространстве ближайшего закрытого родительского слоя </summary>
    property Rect: PRect read GetRect;
    property NodeKey: PWideChar read GetNodeKey;
    property View: TObject read GetView write SetView;
  end;

  /// <summary> Единица в сценарии способная иметь связи </summary>
  IDataStreamProcessor = interface(IScenarioNode)
  ['{B472FC11-DBA9-4FEB-80C3-45996387F0E7}']
    function GetOutputConnections(AIndex: Integer): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}; stdcall;
    procedure SetOutputConnections(AIndex: Integer; const APipe: {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif}); stdcall;
    function GetCountOutputConnections: Integer; stdcall;
    function GetCountInputConnections: Integer; stdcall;
    function GetInputConnections(AIndex: Integer): {$ifdef NARI} Pointer {$else} IScenarioConnection; {$endif} stdcall;
    procedure AddInputConnection(const AConnection: IScenarioConnection); stdcall;
    procedure AddOutputConnection(const AConnection: IScenarioConnection; AAsNumber: Integer); stdcall;
    procedure RemoveInputConnection(const AInputConnection: IScenarioConnection); stdcall;
    procedure RemoveOutputConnection(const AOutputConnection: IScenarioConnection); stdcall;
    function CreateOutputConnection(ANumberConnection, ANumberInput: Integer; const ADestination: IDataStreamProcessor): {$ifdef NARI} Pointer {$else} IScenarioConnection; {$endif} stdcall;
    function HasOutputConnection(const ADestination: IDataStreamProcessor): {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif};
    function GetCountInputDataStreams: Int64; stdcall;
    function GetCountOutputDataStreams: Int64; stdcall;
    function GetInputSpeed: UInt64; stdcall;
    function GetOutputSpeed: UInt64; stdcall;
    procedure SetCountInputDataStreams(const Value: Int64); stdcall;
    procedure SetCountOutputDataStreams(const Value: Int64); stdcall;
    procedure SetInputSpeed(const Value: UInt64); stdcall;
    procedure SetOutputSpeed(const Value: UInt64); stdcall;

    function GetCountLostBlocks: Integer; stdcall;
    function GetCountLostData: Int64; stdcall;

    function CreateDataStream(AParentStream: TDataStream; AInputNumber: Integer): TDataStream; stdcall;
    procedure ProcessData(ADataStream: TDataStream; AData: PByte; ASize: Int64); stdcall;
    procedure DestroyDataStream(ADataStream: TDataStream); stdcall;

    function GetHasInputQueues: boolean; stdcall;
    function GetQueuesItemsCount: UInt32; stdcall;
    function GetQueuesSize: UInt32; stdcall;
    procedure SetQueuesItemsCount(const Value: UInt32); stdcall;
    procedure SetQueuesSize(const Value: UInt32); stdcall;

    property OutputConnections[AIndex: Integer]: {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif} read GetOutputConnections write SetOutputConnections;
    property CountOutputConnections: Integer read GetCountOutputConnections;
    property InputConnections[AIndex: Integer]: {$ifdef NARI} Pointer {$else} IScenarioConnection {$endif} read GetInputConnections;
    property CountInputConnections: Integer read GetCountInputConnections;

    /// данные времени выполнения (не очень наверное что они здесь, пока влом делить на сотни классов и их инкапсуляций ради красивой бэст практис)
    /// <summary> кол-во входных виртуальных потоков данных </summary>
    property CountInputDataStreams: Int64 read GetCountInputDataStreams write SetCountInputDataStreams;
    /// <summary> кол-во вЫходных виртуальных потоков данных </summary>
    property CountOutputDataStreams: Int64 read GetCountOutputDataStreams write SetCountOutputDataStreams;

    /// <summary> входная скорость данных </summary>
    property InputSpeed: UInt64 read GetInputSpeed write SetInputSpeed;
    /// <summary> выходная скорость данных </summary>
    property OutputSpeed: UInt64 read GetOutputSpeed write SetOutputSpeed;

    /// <summary> элемент содержит входную очередь? </summary>
    property HasInputQueues: boolean read GetHasInputQueues;
    /// <summary> колличество элементов в очереди </summary>
    property QueuesItemsCount: UInt32 read GetQueuesItemsCount write SetQueuesItemsCount;
    /// <summary> объём данных в очереди </summary>
    property QueuesSize: UInt32 read GetQueuesSize write SetQueuesSize;

    property CountLostBlocks: Integer read GetCountLostBlocks;
    property CountLostData: Int64 read GetCountLostData;
  end;

  /// <summary> Плагин реализуемыей в библиотеке; одна библиотека может содержать множество плагинов </summary>
  IScenarioPlugin = interface(IDataStreamProcessor)
  ['{5D80DBAD-3AB4-4F1A-AB0B-32B7CDBC62E2}']
    function GetCountTables: Integer; stdcall;
    function GetTable(AIndex: Integer):  {$ifdef NARI} Pointer {$else} IPluginTable {$endif}; stdcall;
    function GetName: PWideChar; stdcall;
    procedure SetName(const AName: PWideChar); stdcall;
    function GetConfiguration: PWideChar; stdcall;
    procedure SetConfiguration(const AValue: PWideChar); stdcall;
    function GetPluginDescription: PPluginDescription; stdcall;
    function GetLoadBeforeStartScenario: boolean; stdcall;
    procedure SetLoadBeforeStartScenario(AValue: boolean); stdcall;
    function GetAsStart: boolean; stdcall;
    procedure SetAsStart(AValue: boolean); stdcall;
    function GetCinemaPlugin: {$ifdef NARI} Pointer {$else} ICinemaPlugin {$endif}; stdcall;

    property Name: PWideChar read GetName write SetName;
    property AsStart: boolean read GetAsStart write SetAsStart;
    property Configuration: PWideChar read GetConfiguration write SetConfiguration;
    property CountTables: Integer read GetCountTables;
    property Table[AIndex: Integer]: {$ifdef NARI} Pointer {$else} IPluginTable {$endif} read GetTable;
    property PluginDescription: PPluginDescription read GetPluginDescription;
    property LoadBeforeStartScenario: boolean read GetLoadBeforeStartScenario write SetLoadBeforeStartScenario;
    property CinemaPlugin: {$ifdef NARI} Pointer {$else} ICinemaPlugin {$endif} read GetCinemaPlugin;
  end;

  /// <summary> Слой сценария. Несёт на себе плагины со связями, а так же другие дочерние слои.
  /// Может быть закрытым (TODO), иметь связи с другими слоями (TODO), выполняться в отдельном потоке </summary>
  IScenarioLayer = interface(IDataStreamProcessor)
  ['{4A055B86-24A9-47AB-B77C-3C6A7641AD38}']
    procedure Lock; stdcall;
    function TryLock: boolean; stdcall;
    procedure Unlock; stdcall;
    function GetChildren(AIndex: Integer): {$ifdef NARI} Pointer {$else} IScenarioNode {$endif}; stdcall;
    function GetCountChildren: Integer; stdcall;
    function GetRectChilds: PRect; stdcall;
    function GetIsThread: Boolean; stdcall;
    procedure SetIsThread(const AValue: Boolean); stdcall;
    function GetIsThreadTerminated: boolean; stdcall;
    function GetIsOpened: boolean; stdcall;
    procedure SetIsOpened(const AValue: boolean); stdcall;
    function GetChildByGuid(const AGuid: TGuid): {$ifdef NARI} Pointer {$else} IScenarioNode; {$endif} stdcall;
    function GetChildrenAtPoint(X, Y: Integer; AIncludOpenedLayers: boolean): {$ifdef NARI} Pointer {$else} IScenarioNode; {$endif} stdcall;
    function GetPipeAtPoint(X, Y: Integer): {$ifdef NARI} Pointer {$else} IScenarioConnection; {$endif} stdcall;
    function CreateChild(const ANodeType: TGUID): {$ifdef NARI} Pointer {$else} IScenarioNode; {$endif} stdcall;
    procedure Clear; stdcall;

    ///  <summary> опрос входных очередей </summary>
    ///  <return> true если хотя бы один элемент очереди был считан </return>
    function AskQueues: boolean; stdcall;
    /// <summary> запуск слоя на самостоятельное выполнение при условии если
    /// он является отдельным потоком и содержит плагины не имеющие входов (стартовые по сценарию) </summary>
    function TryStart: boolean; stdcall;
    function TryStop: boolean; stdcall;
    function TryTerminate: boolean; stdcall;

    function GetLog: {$ifdef NARI} Pointer {$else} ICinemaLog {$endif}; stdcall;

    property Children[AIndex: Integer]: {$ifdef NARI} Pointer {$else} IScenarioNode {$endif} read GetChildren;
    property CountChildren: Integer read GetCountChildren;
    /// <summary> пространство занимаемое дочерними объектами </summary>
    property RectChilds: PRect read GetRectChilds;
    property IsThread: boolean read GetIsThread write SetIsThread;
    property IsThreadTerminated: boolean read GetIsThreadTerminated;

    ///  <summary> открытый блок, т.е. отображается на схеме всегда вмесе с дочерними объектами </summary>
    property IsOpened: boolean read GetIsOpened write SetIsOpened;
    property Log: {$ifdef NARI} Pointer {$else} ICinemaLog {$endif} read GetLog;
  end;

  TEventListener = class abstract
  public
    procedure OnChangeNodeState(const ASourceNode: IScenarioNode); virtual; stdcall; abstract;
    procedure OnChangeLog(const ASourceNode: IScenarioNode); virtual; stdcall; abstract;
  end;

  IStateEvent = interface
    function GetSource: PWideChar; stdcall;
    function GetSourceId: System.PGuid; stdcall;
    function GetState: Integer; stdcall;
    function GetHealthState: Integer; stdcall;
    function GetHealthStateMessage: PWideChar; stdcall;

    property Source: PWideChar read GetSource;
    property SourceId: System.PGuid read GetSourceId;
    property State: Integer read GetState;
    property HealthState: Integer read GetHealthState;
    property HealthStateMessage: PWideChar read GetHealthStateMessage;
  end;

  IMessageEvent = interface
    function GetSource: PWideChar; stdcall;
    function GetSourceId: System.PGuid; stdcall;
    function GetMsg: PWideChar; stdcall;
    function GetEventType: Integer; stdcall;

    property Msg: PWideChar read GetMsg;
    property EventType: int32 read GetEventType;
    property Source: PWideChar read GetSource;
    property SourceId: System.PGuid read GetSourceId;
  end;

  IExternalEventListener = interface
  ['{0DCE2485-6EF2-45FD-BCBD-5FE6405FF580}']
    procedure OnChangeNodeStateEvent(AEvent: IStateEvent); stdcall;
    procedure OnAcceptNewMessageEvent(AEvent: IMessageEvent); stdcall;
  end;

  /// <summary> Сценарий выполняемый ядром. Запуск осуществляется с поиска в слоях-потоках плагинов без входных связей
  /// и последующим созданиям их экземпляров для выполнения работы/алгоритма заложенной автором в его сущность. </summary>
  ICinemaScenario = interface(IScenarioLayer)
  ['{DC2A3B61-8CB5-4875-9805-B3D66C81C220}']
    function LoadScenario(AFileName: PWideChar): boolean; stdcall;
    function SaveScenario(AFileName: PWideChar): boolean; stdcall;
    procedure AddPluginsSearchPath(ASearchPath: PWideChar); stdcall;
    function GetScenarioName: PWideChar; stdcall;
    function GetScenarioFullPath: PWideChar; stdcall;
    function GetEventListener: TEventListener; stdcall;
    procedure SetEventListener(const Value: TEventListener); stdcall;
    function GetExternalEventListener: IExternalEventListener; stdcall;
    procedure SetExternalEventListener(AExternalEventListener: IExternalEventListener); stdcall;
    function GetOnCreateNode: TScenarioNodeEventNotify; stdcall;
    procedure SetOnCreateNode(const AValue: TScenarioNodeEventNotify); stdcall;
    /// <summary> Запрос состояний элементов сценария. Ответ выполнится рабочими потоками через EventListener
    /// если сценарий запущен </summary>
    procedure GetChangedStates; stdcall;
    function GetTagObject: Pointer; stdcall;
    procedure SetTagObject(AValue: Pointer); stdcall;

    /// <summary> Добавить глобальный атрибут, заменяет на новое значение если такой уже есть </summary>
    procedure AddGlobalAttribute(AID: Integer; AData: Pointer; ADataSize: Integer); stdcall;
    function FindGloabalAttribute(AID: Integer): IValueHolder; stdcall;
    procedure RemoveGlobalAttribute(AID: Integer); stdcall;
    procedure AddGlobalNamedAttribute(AName: PWideChar; AData: Pointer; ADataSize: Integer); stdcall;
    function FindGloabalNamedAttribute(AName: PWideChar): IValueHolder; stdcall;
    procedure RemoveGlobalNamedAttribute(AName: PWideChar); stdcall;
    procedure ClearGlobalAttributes; stdcall;

    function GetVersion: PWideChar; stdcall;
    procedure SetVersion(Value: PWideChar); stdcall;

    property ScenarioName: PWideChar read GetScenarioName;
    property ScenarioFullPath: PWideChar read GetScenarioFullPath;
    /// <summary> Через EventListener отправляются события генерируемые элементами сценария.
    ///  Если сценарий выполняется, тогда реализации TEventListener следует позаботиться о синхронизации
    ///  приёма событий, т.к. они отправляются из рабочих потоков </summary>
    property EventListener: TEventListener read GetEventListener write SetEventListener;
    property ExternalEventListener: IExternalEventListener read GetExternalEventListener write SetExternalEventListener;
    property OnCreateNode: TScenarioNodeEventNotify read GetOnCreateNode write SetOnCreateNode;
    /// <summary> Любой сторонний объект с которым ассоциируется сценарий </summary>
    property TagObject: Pointer read GetTagObject write SetTagObject;
    property Version: PWideChar read GetVersion write SetVersion;
  end;

const
  NODE_STATE_STRINGS: array[TNodeState] of string =
  (
    'Не загружен',
    'Запуск',
    'Работает',
    'Нет данных',
    'Остановка'
  );

  OPEN_LAYER_BORDER_SIZE = 10;

  MSG_TYPE_TO_HEALTH_DEGREE: array[TNodeLogState] of TNodeHalthDegreeState = (100, 50, 0);

  csCrossLineBreak: WideString = #$0d#$0a;
  csTab: WideString = #$0009;

implementation

end.
