﻿unit Cinema.Types;

{$ifdef FPC}
  {$mode delphi}
{$endif}

interface

uses

  {$ifdef FPC}
  Types,
  SysUtils,
  fpJson;
  {$else}
  System.Types,
  System.SysUtils,
  System.JSON;
  {$endif}

type

  // Data.DB.TFieldType
  TAttributeType = (atRaw, atString, atSmallint, atInteger, atWord, // 0..4
    atBoolean, atFloat, atCurrency, atBCD, atDate, atTime, atDateTime, // 5..11
    atBytes, atVarBytes, atAutoInc, atBlob, atMemo, atGraphic, atFmtMemo, // 12..18
    atParadoxOle, atDBaseOle, atTypedBinary, atCursor, atFixedChar, atWideString, // 19..24
    atLargeint, atADT, atArray, atReference, atDataSet, atOraBlob, atOraClob, // 25..31
    atVariant, atInterface, atIDispatch, atGuid, atTimeStamp, atFMTBcd, // 32..37
    atFixedWideChar, atWideMemo, atOraTimeStamp, atOraInterval, // 38..41
    atLongWord, atShortint, atByte, atExtended, atConnection, atParams, atStream, //42..48
    atTimeStampOffset, atObject, atSingle);

  TMessageType = Integer;
  TCinemaFieldType = TAttributeType;

  TJsonCinema = {$ifdef FPC} TJSONData {$else} TJSONValue {$endif};

type
  TCinemaSharedPointer<T> = record
  private
    FValue: T;
    FCounter: PInteger;
  public
    constructor Create(const AValue: T);
    class operator Initialize({$ifdef FPC}var{$else}out{$endif} ADest: TCinemaSharedPointer<T>);
    class operator Finalize(var ADest: TCinemaSharedPointer<T>);
    class operator Implicit(const AValue: T): TCinemaSharedPointer<T>;
    class operator Implicit(const ASrc: TCinemaSharedPointer<T>): T;
    {$ifdef FPC}
    class operator Copy(constref ASrc: TCinemaSharedPointer<T>; var ADest: TCinemaSharedPointer<T>);
    {$else}
    class operator Assign(var ADest: TCinemaSharedPointer<T>; const [ref] ASrc: TCinemaSharedPointer<T>);
    {$endif}
    property Value: T read FValue;
  end;


const

  MESSAGE_INFO:  TMessageType = 0;
  MESSAGE_WARN:  TMessageType = 1;
  MESSAGE_ERROR: TMessageType = 2;

  HEALTH_STATE_OK = 100; // %
  HEALTH_STATE_DEAD = 0; // %

{$ifdef MSWINDOWS}
  ContrPathDelim = '/';
{$else}
  ContrPathDelim = '\';
{$endif}

const

{$ifdef win32}
  _PLATFORM = 'Win32';
  csCinemaCore = 'Cinema.dll';
{$endif}
{$ifdef win64}
  _PLATFORM = 'Win64';
  csCinemaCore = 'Cinema.dll';
{$endif}
{$ifdef linux}
  csCinemaCore = 'libcinema.so';
  {$ifdef CPUX64}
  _PLATFORM = 'Linux64';
  {$endif}
  {$ifdef CPUX32}
  _PLATFORM = 'Linux32';
  {$endif}
  {$ifdef ARM32}
  _PLATFORM = 'ARM32';
  {$endif}
  {$ifdef ARM64}
  _PLATFORM = 'ARM64';
  {$endif}
{$endif}
    csPluginsDir = 'Plugins';
    csScenarioDir = 'Scenario';
    csWorkDir: UnicodeString = 'work' + PathDelim;
    csWorkSubDirScens: UnicodeString = 'scenarios' + PathDelim;

implementation

uses
  {$ifdef FPC}
  Math,
  {%H-}Rtti,
  TypInfo;
  {$else}
  System.Math,
  System.Rtti,
  System.TypInfo;
  {$endif}

{ TCinemaSharedPointer<T> }

{$ifdef FPC}
class operator TCinemaSharedPointer<T>.Copy(constref ASrc: TCinemaSharedPointer<T>; var ADest: TCinemaSharedPointer<T>);
{$else}
class operator TCinemaSharedPointer<T>.Assign(var ADest: TCinemaSharedPointer<T>; const [ref] ASrc: TCinemaSharedPointer<T>);
{$endif}
begin
  if (ADest.FCounter <> ASrc.FCounter) then
    Finalize(ADest);

  ADest.FValue := ASrc.FValue;
  ADest.FCounter := ASrc.FCounter;
  if Assigned(ADest.FCounter) then
  {$ifdef fpc}
    InterlockedIncrement(ADest.FCounter^);
  {$else}
    AtomicIncrement(ADest.FCounter^);
  {$endif}
end;

class operator TCinemaSharedPointer<T>.Implicit(const AValue: T): TCinemaSharedPointer<T>;
begin
  Result := TCinemaSharedPointer<T>.Create(AValue);
end;

constructor TCinemaSharedPointer<T>.Create(const AValue: T);
begin
  new(FCounter);
  FCounter^ := 1;
  FValue := AValue;
end;

class operator TCinemaSharedPointer<T>.Finalize(var ADest: TCinemaSharedPointer<T>);
var
  io: TInterfacedObject;
begin
  if Assigned(ADest.FCounter) then
  begin
    {$ifdef fpc}
    InterlockedIncrement(ADest.FCounter^);
    {$else}
    AtomicDecrement(ADest.FCounter^);
    {$endif}
    if ADest.FCounter^ = 0 then
    begin
      if PTypeInfo(TypeInfo(T)).Kind = TTypeKind.tkClass then
        TObject(Pointer(@ADest.FValue)^).Free;
      Dispose(ADest.FCounter);
      ADest.FCounter := nil;
    end;
  end;
end;

class operator TCinemaSharedPointer<T>.Implicit(const ASrc: TCinemaSharedPointer<T>): T;
begin
  Result := ASrc.FValue;
end;

class operator TCinemaSharedPointer<T>.Initialize({$ifdef FPC} var {$else} out {$endif} ADest: TCinemaSharedPointer<T>);
begin
  ADest.FCounter := nil;
end;

end.
