﻿unit Cinema.Configurator.View;

{$ifdef FPC}
  {$mode Delphi}
{$endif}

interface

uses
    SysUtils
  , Types
  , System.UITypes
  , Classes
  , Variants
  , Forms
  , Controls
  , StdCtrls
  , Graphics
  , Buttons
  , ActnList
  , ExtCtrls
  , Cinema.Scenario.Interfaces
  , Cinema.Interfaces;

type
  TfrmConfigurator = class(TForm)
    btnCancel: TButton;
    ImageList: TImageList;
    ActionList: TActionList;
    acnAccept: TAction;
    acnCancel: TAction;
    PanelMain: TPanel;
    btnAccept: TButton;
    MemoJson: TMemo;
    sbPluginHelp: TSpeedButton;
    procedure acnAcceptExecute(Sender: TObject);
    procedure acnCancelExecute(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure sbPluginHelpClick(Sender: TObject);
  private
    FPluginDescription: PPluginDescription;
    function SchemeToDefaultConfig(const AScheme: string): string;
  public
    function Configurate(APluginDescription: PPluginDescription; var AJsonModel: string): boolean;
  end;

var
  frmConfigurator: TfrmConfigurator;

implementation

uses
  {$ifdef FPC}
    fpjson
  , jsonparser
  {$else}
    System.JSON
  {$endif}
  , Cinema.Types
  , Cinema.Configurator.View.HelpDialog
  , Cinema.Utils
  ;

{$ifdef FPC}
  {$R *.lfm}
{$else}
  {$R *.fmx}
{$endif}

{ TfrmConfigurator }

procedure TfrmConfigurator.acnAcceptExecute(Sender: TObject);
begin
  Close;
  ModalResult := mrOk;
end;

procedure TfrmConfigurator.acnCancelExecute(Sender: TObject);
begin
  Close;
end;

function TfrmConfigurator.Configurate(APluginDescription: PPluginDescription; var AJsonModel: string): boolean;
var
  jsnSchemeValue: TJSONCinema;
begin
  FPluginDescription := APluginDescription;
  sbPluginHelp.Visible := Assigned(FPluginDescription);

  // временное решение - вывожу джейсон с конфигурацией в мемо

  if AJsonModel.IsEmpty then
    MemoJson.Lines.Text := SchemeToDefaultConfig(APluginDescription.SchemeConfig)
  else
    MemoJson.Lines.Text := AJsonModel;

  Result := ShowModal = mrOk;

  if Result then
    AJsonModel := MemoJson.Lines.Text;

  exit;

  // TODO

  if APluginDescription.SchemeConfig = nil then
    exit;

  jsnSchemeValue := TJSONObject.ParseJSONValue(APluginDescription.SchemeConfig);
  try

    if not Assigned(jsnSchemeValue) then
      exit;

    //for jsnSchemeItem in TJSONArray(jsnSchemeValue) do
    //begin
    //end;

  finally
    jsnSchemeValue.Free;
  end;
end;

procedure TfrmConfigurator.FormShow(Sender: TObject);
begin

  if Left < 0 then
    Left := 0;

  if Top < 0 then
    Top := 0;

end;

procedure TfrmConfigurator.sbPluginHelpClick(Sender: TObject);
begin
  frmConfiguratorHelpDialog.ShowHelpDialog(FPluginDescription);
end;

function TfrmConfigurator.SchemeToDefaultConfig(const AScheme: string): string;

  function ProcessArrayItem(const AIdent: string; AItem: TJSONCinema): string;
  var
    i: Int32;
    val: string;
    name: string;
    default: string;
    incapsulated: TJSONCinema;
    incapItem: TJSONCinema;
    count: int32;
    resOk: boolean;
  begin
    Result := '';
    resOk := AItem.TryGetValue<string>('type', val);
    if resOk then
    begin
      AItem.TryGetValue<string>('name', name);
      if val = 'list' then
      begin
        Result := AIdent + '"' + name + '":[' + sLineBreak;
        incapsulated := AItem.FindPath('value');
        if Assigned(incapsulated) then
        begin
          count := 0;
          Result := Result + AIdent + '  {';
          for i := 0 to TJSONArray(incapsulated).Count - 1 do
          begin
            incapItem := TJSONArray(incapsulated).Items[i];
            if count > 0 then
              Result := Result + ',' + sLineBreak + ProcessArrayItem(AIdent + '    ', incapItem)
            else
              Result := Result + sLineBreak + ProcessArrayItem(AIdent + '    ', incapItem) ;
            inc(count);
          end;
          Result := Result + sLineBreak + AIdent + '  }';
        end;

        Result := Result + sLineBreak + AIdent + ']';
      end else
      begin
        AItem.TryGetValue<string>('default', default);
        if val = 'number' then
        begin
          Result := AIdent + '"' + name + '": ' + default;
        end else
        begin
          Result := AIdent + '"' + name + '": "' + default + '"';
        end;
      end;

    end;
    //Result := Result + '}';
  end;

var
  i: Int32;
  jsnSchemeValue: TJSONCinema;
  jsnSchemeItem: TJSONCinema;
  count: int32;
begin
  if AScheme.IsEmpty then
    exit('');

  // соберём настройки по умолчанию из спецификации
  jsnSchemeValue := TJSONObject.ParseJSONValue(StringToWide(AScheme));
  try

    if not Assigned(jsnSchemeValue) then
      exit('');

    Result := '{';
    count := 0;
    for i := 0 to TJSONArray(jsnSchemeValue).Count - 1 do
    begin
      jsnSchemeItem := TJSONArray(jsnSchemeValue).Items[i];
      if count > 0 then
        Result := Result + ',' + sLineBreak + ProcessArrayItem('  ', jsnSchemeItem)
      else
        Result := Result  + sLineBreak + ProcessArrayItem('  ', jsnSchemeItem);
      inc(count);
    end;
    Result := Result + sLineBreak + '}'

  finally
    jsnSchemeValue.Free;
  end;
end;

end.
