﻿unit Cinema.Scenario.Statistics;

{$ifdef fpc}
  {$mode Delphi}
{$endif}

interface

uses
  {$ifndef fpc}
  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants, 
  FMX.Types, FMX.Graphics, FMX.Controls, FMX.Forms, FMX.Dialogs, FMX.StdCtrls, System.Rtti, FMX.Grid.Style, FMX.Grid, FMX.Controls.Presentation, FMX.ScrollBox,
  FMX.Layouts, FMX.TabControl, FMX.Objects, FMX.Menus, FMX.Edit, System.ImageList, FMX.ImgList,
  {$else}
    SysUtils
  , Generics.Collections
  , Types
  , Controls
  , Forms
  , StdCtrls
  , Classes
  , Variants
  , Graphics
  , ExtCtrls
  , ActnList
  , ComCtrls
  , Buttons
  , Menus
  , Grids, EditBtn
  {$endif}

  , Cinema.Interfaces
  , Cinema.Scenario.Interfaces
  , Cinema.Hypervisor.Interfaces
  ;

type

  TStatType = (stProcessor, stScenario, stCinemaNode, stHypervisor, stNode);

  { TFrameStatistic }

  TFrameStatistic = class(TFrame)
    SearchEdit: TEditButton;
    LayoutScenarioPluginStat: TPanel;
    loComParamName: TPanel;
    miCopyRow: TMenuItem;
    miCopyCell: TMenuItem;
    Splitter1: TSplitter;
    lblState: TLabel;
    loCommParamValue: TPanel;
    lblStateValue: TLabel;
    lblHealthState: TLabel;
    lblHealthStateValue: TLabel;
    lblDataStreams: TLabel;
    lblDataStreamsValues: TLabel;
    SplitterCommonAndCustom: TSplitter;
    lblSpeed: TLabel;
    lblSpeedValue: TLabel;
    LayoutScenarioStat: TPanel;
    Layout2: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Splitter2: TSplitter;
    Layout3: TPanel;
    lblScenState: TLabel;
    lblScenHealth: TLabel;
    lblScenMemory: TLabel;
    lblScenCPU: TLabel;
    LayoutCinemaNodeStat: TPanel;
    Layout5: TPanel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Splitter3: TSplitter;
    Layout6: TPanel;
    lblHostName: TLabel;
    lblHostAddress: TLabel;
    lblHostMemory: TLabel;
    lblHostCPU: TLabel;
    Label9: TLabel;
    lblHostIsConnected: TLabel;
    LayoutHypervisor: TPanel;
    Layout4: TPanel;
    Label10: TLabel;
    Label12: TLabel;
    Label13: TLabel;
    Label14: TLabel;
    Splitter4: TSplitter;
    Layout7: TPanel;
    lblHypervisorName: TLabel;
    lblHypervisorMem: TLabel;
    lblHypervisorCPU: TLabel;
    lblHypervisorNodes: TLabel;
    Label11: TLabel;
    lblVersion: TLabel;
    LayoutNodeStat: TPanel;
    Layout8: TPanel;
    Label15: TLabel;
    Label17: TLabel;
    Label18: TLabel;
    Splitter5: TSplitter;
    Layout9: TPanel;
    lblCountChildrenNode: TLabel;
    lblUsageMemNode: TLabel;
    lblCpuUsageNode: TLabel;
    Label23: TLabel;
    lblAvailableMemNode: TLabel;
    lblPopupMenu: TPopupMenu;
    miCopyMessage: TMenuItem;
    miCopyTable: TMenuItem;
    PopupMenuTable: TPopupMenu;
    LayoutTables: TPanel;
    tcTables: TPageControl;
    ImageList: TImageList;
    lblQueueSizeValue: TLabel;
    lblQueueSize: TLabel;
    procedure miCopyCellClick(Sender: TObject);
    procedure miCopyMessageClick(Sender: TObject);
    procedure lblStateValueMouseEnter(Sender: TObject);
    procedure miCopyRowClick(Sender: TObject);
    procedure miCopyTableClick(Sender: TObject);
    procedure SearchEditButtonClick(Sender: TObject);
    procedure SearchEditKeyUp(Sender: TObject; var Key: Word; var KeyChar: Char; Shift: TShiftState);
  private
    class var FTotalMemory: int64;
    class var FSysTime: TThread.TSystemTimes;
  private

    FStatType: TStatType;
    FTextClick: string;
    FColSelected: Int32;
    FUserAdjusted: boolean;
    FLastUpdatedProcessorStat: IDataStreamProcessor;
    FGridLoading: TStringGrid;

    procedure LoadCustomTable(AGrid: TStringGrid; const ATable: IPluginTable);
    procedure SetStatType(const Value: TStatType);
                   //procedure(Sender: TObject; IsColumn: Boolean; Index: Integer)
    procedure OnHeaderSized(ASender: TObject; IsColumn: Boolean; Index: Integer);
    procedure DoUpdateScenarioStat(const AScenario: ICinemaScenario);
    class constructor Create;
  public
    // TODO - общая статистика по сценарию, хосту
    procedure UpdateProcessorStat(const AProcessor: IDataStreamProcessor);
    procedure UpdateScenarioStat(const AScenario: IHypervisorScenario); overload;
    procedure UpdateScenarioStat(const AScenario: ICinemaScenario); overload;
    procedure UpdateCinemaNodeStat(const ASourceNode: IHypervisorCinemaNode);
    procedure UpdateHypervisorStat(const ASource: IHypervisor);
    procedure UpdateNodeStat(const ASourceNode: IHypervisorNode);

    property StatType: TStatType read FStatType write SetStatType;
  end;

implementation

uses
    StrUtils
  {$ifdef FPC}
  , LCLType
  , Clipbrd
  {$else}
  , FMX.Clipboard,
  , FMX.Platform,
  {$endif}
  , Cinema.Utils
  ;

{$ifdef FPC}
  {$R *.lfm}
{$else}
  {$R *.fmx}
{$endif}

const
  // копия из Cinema.Scenario.Interfaces, т.к. компилятор из другого юнита неверно перекодирует в UTF8
  NODE_STATE_STRINGS: array[TNodeState] of Unicodestring =
  (
    'Не загружен',
    'Запуск',
    'Работает',
    'Нет данных',
    'Остановка'
  );

function BytesToString(ABytes: int64): string;
const
  KILOBYTE = 1000;
  MEGABYTE = KILOBYTE * 1000;
  GIGABYTE = MEGABYTE * 1000;
begin
  if ABytes > GIGABYTE then
    Result := Format('%f', [ABytes/GIGABYTE]) + ' G'
  else
  if ABytes > MEGABYTE then
    Result := Format('%f', [ABytes/MEGABYTE]) + ' M'
  else
  if ABytes > KILOBYTE then
    Result := Format('%f', [ABytes/KILOBYTE]) + ' K'
  else
    Result := IntToStr(ABytes) + ' B';
end;

{ TFrameStatistic }

class constructor TFrameStatistic.Create;
begin
  TThread.GetSystemTimes(TFrameStatistic.FSysTime);
  TFrameStatistic.FTotalMemory := GetTotalMemory;
end;

procedure TFrameStatistic.DoUpdateScenarioStat(const AScenario: ICinemaScenario);
var
  healthState: string;
begin

  if FStatType <> stScenario then
    StatType := stScenario;

  lblScenState.Caption := WideToString(NODE_STATE_STRINGS[AScenario.State]);

  if AScenario.State <> TNodeState.nsNone then
  begin
    healthState := WideToString(AScenario.HealthStateMessage);
    if healthState.IsEmpty then
      lblScenHealth.Caption := IntToStr(AScenario.HealthState) + ' %'
    else
      lblScenHealth.Caption := healthState + ' (' + IntToStr(AScenario.HealthState) + ' %)';
  end else
  begin
    lblScenHealth.Caption := WideToString(NODE_STATE_STRINGS[TNodeState.nsNone]);
  end;
end;

procedure TFrameStatistic.lblStateValueMouseEnter(Sender: TObject);
begin
  if Sender.InheritsFrom(TLabel) then
    FTextClick := TLabel(Sender).Caption;
end;

procedure TFrameStatistic.miCopyRowClick(Sender: TObject);
var
  i: int32;
  row: string;
  strGrid: TStringGrid;
  clip: TClipboard;
begin
  if (tcTables.TabIndex < 0) then
    exit;

  strGrid := TStringGrid(tcTables.Pages[tcTables.TabIndex].Tag);
  if not Assigned(strGrid) or (strGrid.RowCount = 0) then
    exit;

  if (strGrid.Row >= 0) then
  begin
    row := '';
    for i := 0 to strGrid.ColCount - 1 do
      if i > 0 then
        row := row + ';' + strGrid.Cells[i, strGrid.Row]
      else
        row := strGrid.Cells[i, strGrid.Row];
    clip := Clipboard(TClipboardType.ctClipboard);
    clip.SetTextBuf(PChar(row));
  end;
end;

procedure TFrameStatistic.LoadCustomTable(AGrid: TStringGrid; const ATable: IPluginTable);
const
  csMaxColumnWidth = 300;
var
  i, j: int32;
  textWidth: Int32;
  column: TGridColumn;
  cellText: string;
  colWidth: Int32;
begin
  AGrid.BeginUpdate;
  try
    FGridLoading := AGrid;
    AGrid.RowCount := ATable.CountRows + AGrid.FixedRows;
    if AGrid.FixedCols <> 0 then
      AGrid.FixedCols := 0;
    for i := 0 to ATable.CountColumn - 1 do
    begin
      if AGrid.Columns.Count > i then
      begin
        column := AGrid.Columns.Items[i];
      end else
      begin
        column := AGrid.Columns.Add;
      end;
      column.Title.Caption := ATable.GetColumnCaption(i);
      if not FUserAdjusted then
        column.Width := AGrid.Canvas.TextWidth(column.Title.Caption) + 8;
      colWidth := column.Width;
      for j := 0 to ATable.CountRows - 1 do
      begin
        //if j + AGrid.FixedRows >= AGrid.RowCount then
        //  AGrid.RowCount := ATable.CountRows;
        cellText := WideToString(ATable.GetCell(i, j));
        AGrid.Cells[i, AGrid.FixedRows + j] := cellText;
        if not FUserAdjusted then
        begin
          textWidth := AGrid.Canvas.TextWidth(cellText);
          if colWidth < textWidth + 4 then
            colWidth := textWidth + 4;
        end;
      end;

      if not FUserAdjusted and (colWidth < csMaxColumnWidth) then
        column.Width := colWidth;
    end;

    // если вдруг оказались лишние колонки - удаляем
    while AGrid.Columns.Count > ATable.CountColumn do
    begin
      AGrid.Columns.Delete(AGrid.Columns.Count - 1);
    end;
  finally
    AGrid.EndUpdate;
    FGridLoading := nil;
  end;

end;

procedure TFrameStatistic.miCopyMessageClick(Sender: TObject);
var
  clip: TClipboard;
begin
  clip := Clipboard(TClipboardType.ctClipboard);
  clip.SetTextBuf(PChar(FTextClick));
end;

procedure TFrameStatistic.miCopyCellClick(Sender: TObject);
var
  strGrid: TStringGrid;
  clip: TClipboard;
begin
  if (tcTables.TabIndex < 0) then
    exit;

  strGrid := TStringGrid(tcTables.Pages[tcTables.TabIndex].Tag);
  if not Assigned(strGrid) or (strGrid.RowCount = 0) then
    exit;

  if (strGrid.Row >= 0) and (strGrid.Col >= 0) then
  begin
    clip := Clipboard(TClipboardType.ctClipboard);
    clip.SetTextBuf(PChar(strGrid.Cells[strGrid.Col, strGrid.Row]));
  end;
end;

procedure TFrameStatistic.miCopyTableClick(Sender: TObject);
var
  clip: TClipboard;
  strGrid: TStringGrid;
  result: string;
  i, j: Integer;
begin
  if tcTables.TabIndex < 0 then
    exit;

  strGrid := TStringGrid(tcTables.Pages[tcTables.TabIndex].Tag);
  if not Assigned(strGrid) then
    exit;

  result := '';

  for i := 0 to strGrid.Columns.Count - 1 do
    if i > 0 then
      result := result + #9 + strGrid.Columns[i].Title.Caption
    else
      result := strGrid.Columns[i].Title.Caption;

  result := result + sLineBreak;

  for i := 0 to strGrid.RowCount - 1 do
  begin
    for j := 0 to strGrid.Columns.Count - 1 do
      if j > 0 then
        result := result + #9 + strGrid.Cells[j, i]
      else
        result := result + strGrid.Cells[j, i];

    result := result + sLineBreak;
  end;


  clip := Clipboard(TClipboardType.ctClipboard);
  clip.SetTextBuf(PChar(result));

end;

procedure TFrameStatistic.OnHeaderSized(ASender: TObject; IsColumn: Boolean; Index: Integer);
begin
  if Assigned(FGridLoading) then
    exit;
  FUserAdjusted := true;
end;

procedure TFrameStatistic.SearchEditButtonClick(Sender: TObject);
var
  strGrid: TStringGrid;
  searchedText: string;
  rowIndex, endIndex, j: Integer;
begin
  if (tcTables.TabIndex < 0) or (SearchEdit.Text = '') then
    exit;

  strGrid := TStringGrid(tcTables.Pages[tcTables.TabIndex].Tag);
  if not Assigned(strGrid) or (strGrid.RowCount = 0) then
    exit;

  searchedText := UpperCase(SearchEdit.Text);

  if strGrid.Row >= 0 then
  begin
    endIndex := strGrid.Row;
    rowIndex := strGrid.Row;
  end else
  begin
    rowIndex := 0;
    endIndex := strGrid.RowCount;
  end;

  j := FColSelected + 1;
  repeat
    while j < strGrid.Columns.Count do
    begin
      if UpperCase(strGrid.Cells[j, rowIndex]).Contains(searchedText) then
      begin
        FColSelected := j;
        strGrid.Row := rowIndex;
        strGrid.Col := j;
        //strGrid.ScrollToSelectedCell;
        //strGrid.SetFocus;
        exit;
      end;
      inc(j);
    end;
    j := 0;
    inc(rowIndex);
    if rowIndex >= strGrid.RowCount then
    begin
      if rowIndex = endIndex then
        break;
      rowIndex := 0;
    end;

  until rowIndex = endIndex;

end;

procedure TFrameStatistic.SearchEditKeyUp(Sender: TObject; var Key: Word; var KeyChar: Char; Shift: TShiftState);
begin
  if Key = VK_RETURN then
    SearchEditButtonClick(Sender);
end;

procedure TFrameStatistic.SetStatType(const Value: TStatType);
begin
  if FStatType = Value then
    exit;

  FStatType := Value;
  case FStatType of
    stProcessor: begin
      LayoutCinemaNodeStat.Visible := false;
      LayoutScenarioStat.Visible := false;
      LayoutNodeStat.Visible := false;
      LayoutScenarioPluginStat.Visible := true;
      SplitterCommonAndCustom.Visible := true;
      LayoutHypervisor.Visible := false;
      tcTables.Visible := true;
    end;

    stScenario: begin
      LayoutScenarioPluginStat.Visible := false;
      tcTables.Visible := false;
      LayoutNodeStat.Visible := false;
      SplitterCommonAndCustom.Visible := false;
      LayoutCinemaNodeStat.Visible := false;
      LayoutHypervisor.Visible := false;
      LayoutScenarioStat.Visible := true;
    end;

    stCinemaNode: begin
      LayoutScenarioStat.Visible := false;
      LayoutScenarioPluginStat.Visible := false;
      LayoutNodeStat.Visible := false;
      tcTables.Visible := false;
      SplitterCommonAndCustom.Visible := false;
      LayoutHypervisor.Visible := false;
      LayoutCinemaNodeStat.Visible := true;
    end;

    stHypervisor: begin
      LayoutScenarioStat.Visible := false;
      LayoutScenarioPluginStat.Visible := false;
      LayoutNodeStat.Visible := false;
      tcTables.Visible := false;
      SplitterCommonAndCustom.Visible := false;
      LayoutCinemaNodeStat.Visible := false;
      LayoutHypervisor.Visible := true;
    end;

    stNode: begin
      LayoutScenarioStat.Visible := false;
      LayoutScenarioPluginStat.Visible := false;
      tcTables.Visible := false;
      SplitterCommonAndCustom.Visible := false;
      LayoutCinemaNodeStat.Visible := false;
      LayoutHypervisor.Visible := false;
      LayoutNodeStat.Visible := true;
    end;
  end;
  SearchEdit.Visible := tcTables.Visible;
end;

procedure TFrameStatistic.UpdateHypervisorStat(const ASource: IHypervisor);
var
  availableMem: int64;
begin
  if FStatType <> stHypervisor then
    StatType := stHypervisor;

  lblHypervisorName.Caption := WideToString(ASource.Name);
  lblHypervisorNodes.Caption := IntToStr(ASource.CountConnectedCinemaNodes) + ' / ' + IntToStr(ASource.CountCinemaNodes);

  availableMem := GetAvailableMemory;
  lblHypervisorMem.Caption := IntToStr(round(availableMem/FTotalMemory*100)) + ' % / ' + BytesToString(availableMem);

  lblHypervisorCPU.Caption := IntToStr(TThread.GetCPUUsage(FSysTime)) + ' %';

end;

procedure TFrameStatistic.UpdateNodeStat(const ASourceNode: IHypervisorNode);
var
  sumStat: TSumResources;
begin
  if FStatType <> stNode then
    StatType := stNode;

  sumStat := ASourceNode.SumResources;
  lblCountChildrenNode.Caption := IntToStr(ASourceNode.CountChildren);
  if sumStat.TotalMemory > 0 then
    lblAvailableMemNode.Caption := IntToStr(round((sumStat.TotalMemory - sumStat.UsageMemory)/sumStat.TotalMemory*100)) + string(' % / ')
      + BytesToString(sumStat.TotalMemory - sumStat.UsageMemory) + string(' из ') + BytesToString(sumStat.TotalMemory)
  else
    lblAvailableMemNode.Caption := '0 % / 0';

  lblUsageMemNode.Caption := BytesToString(sumStat.UsageMemory);
  lblCpuUsageNode.Caption := IntToStr(sumStat.CPUUsage) + ' %';
 end;

procedure TFrameStatistic.UpdateCinemaNodeStat(const ASourceNode: IHypervisorCinemaNode);
begin
  if FStatType <> stCinemaNode then
    StatType := stCinemaNode;


  lblHostName.Caption := WideToString(ASourceNode.Name);
  lblHostAddress.Caption := ASourceNode.Address;
  if ASourceNode.TotalMemory > 0 then
    lblHostMemory.Caption := IntToStr(round((ASourceNode.TotalMemory - ASourceNode.UsageMemory)/ASourceNode.TotalMemory*100)) + string(' % / ')
      + BytesToString(ASourceNode.TotalMemory - ASourceNode.UsageMemory) + string(' из ') + BytesToString(ASourceNode.TotalMemory)
  else
    lblHostMemory.Caption := '0 % / 0';
  lblHostCPU.Caption := IntToStr(ASourceNode.CPUUsage) + ' %';
  lblHostIsConnected.Caption := IfThen(ASourceNode.IsConnected, 'Да', 'Нет');
  lblVersion.Caption := WideToString(ASourceNode.Version);

end;

procedure TFrameStatistic.UpdateProcessorStat(const AProcessor: IDataStreamProcessor);
var
  i, countTables: int32;
  healthState: string;
  plugin: IScenarioPlugin;
  table: TStringGrid;
  tab: TTabSheet;
begin
  if FStatType <> stProcessor then
    StatType := stProcessor;

  if FLastUpdatedProcessorStat <> AProcessor then
  begin
    FLastUpdatedProcessorStat := AProcessor;
    FColSelected := -1;
    FUserAdjusted := false;
    lblQueueSize.Visible := AProcessor.HasInputQueues;
    lblQueueSizeValue.Visible := lblQueueSize.Visible;
  end;

  lblStateValue.Caption := WideToString(NODE_STATE_STRINGS[AProcessor.State]);
  if AProcessor.State <> TNodeState.nsNone then
  begin
    healthState := AProcessor.HealthStateMessage;
    if healthState.IsEmpty then
      lblHealthStateValue.Caption := IntToStr(AProcessor.HealthState) + '%'
    else
      lblHealthStateValue.Caption := healthState;
  end else
  begin
    lblHealthStateValue.Caption := WideToString(NODE_STATE_STRINGS[TNodeState.nsNone]);
  end;

  lblDataStreamsValues.Caption := IntToStr(AProcessor.CountInputDataStreams) + '/' + IntToStr(AProcessor.CountOutputDataStreams);
  lblSpeedValue.Caption := BytesToString(AProcessor.InputSpeed) + ' / ' + BytesToString(AProcessor.OutputSpeed);

  if lblQueueSizeValue.Visible then
  begin
    lblQueueSizeValue.Caption := IntToStr(AProcessor.QueuesItemsCount) + '/' + BytesToString(AProcessor.QueuesSize);
  end;

  countTables := 0;

  if (AProcessor.QueryInterface(IScenarioPlugin, plugin) = S_OK) then
  begin
    tcTables.Visible := plugin.CountTables > 0;
    SplitterCommonAndCustom.Visible := tcTables.Visible;
    for i := 0 to plugin.CountTables - 1 do
    begin
      inc(countTables);
      if tcTables.PageCount <= i then
      begin
        tab := tcTables.AddTabSheet;
        table := TStringGrid.Create(tab);
        table.PopupMenu := PopupMenuTable;
        table.Parent := tab;
        tab.Tag := PtrInt(table);
        table.Align := TAlign.alClient;
        table.Options := table.Options + [goFixedVertLine,goFixedHorzLine,goVertLine,goHorzLine,goRangeSelect,goDrawFocusSelected,goColSizing,goEditing,goSmoothScroll];
        table.OnHeaderSized := OnHeaderSized;
      end else
      begin
        tab := tcTables.Pages[i];
        table := TStringGrid(tab.Tag);
      end;
      tab.Caption := IPluginTable(plugin.Table[i]).Caption;
      LoadCustomTable(table, IPluginTable(plugin.Table[i]));
    end;
  end else
  begin
    tcTables.Visible := false;
    SplitterCommonAndCustom.Visible := false;
  end;
  SearchEdit.Visible := tcTables.Visible;

  while tcTables.PageCount > countTables do
  begin
    tcTables.Pages[tcTables.PageCount - 1].Free;
  end;
end;

procedure TFrameStatistic.UpdateScenarioStat(const AScenario: ICinemaScenario);
var
  usageMem: Int64;
begin
  DoUpdateScenarioStat(AScenario);

  if FTotalMemory > 0 then
  begin
    usageMem := GetWorkingSetSize;
    lblScenMemory.Caption := BytesToString(usageMem) + ' / ' + IntToStr(Round(usageMem/FTotalMemory*100)) + ' %';
  end else
    lblScenMemory.Caption := '0 / 0 %';

  lblScenCPU.Caption := IntToStr(TThread.GetCPUUsage(FSysTime)) + ' %';
end;

procedure TFrameStatistic.UpdateScenarioStat(const AScenario: IHypervisorScenario);
var
  scen: ICinemaScenario;
begin

  scen := AScenario.Scenario;
  DoUpdateScenarioStat(scen);

  if AScenario.TotalMemory > 0 then
    lblScenMemory.Caption := BytesToString(AScenario.UsageMemory) + ' / ' + IntToStr(Round(AScenario.UsageMemory/AScenario.TotalMemory*100)) + ' %'
  else
    lblScenMemory.Caption := '0 / 0 %';

  lblScenCPU.Caption := IntToStr(AScenario.CPUUsage) + ' %';

end;

end.
